﻿#region Copyright (c) 2013-04, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Reflection;
using Amarok.Agents.Resources;


namespace Amarok.Agents
{
	/// <summary>
	/// This type implements the default factory that is capable to spawn new agent instances.
	/// </summary>
	public class DefaultAgentFactory :
		IAgentFactory
	{
		#region ++ IAgentFactory Interface ++

		/// <summary>
		/// Spawns a new instance of the specified agent type.
		/// </summary>
		/// 
		/// <param name="agentType">
		/// The type of agent to spawn. This must be a sub-class of <see cref="Agent"/>.</param>
		/// <param name="environment">
		/// The agent environment to be supplied to the new agent instance.</param>
		/// <param name="options">
		/// The agent options to be supplied to the new agent instance.</param>
		/// 
		/// <returns>
		/// A reference to the newly spawned agent instance.</returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentException">
		/// A value was passed to a method that did not accept it as a valid argument, because 
		/// the value representing a type must be a sub-class of <see cref="Agent"/>.</exception>
		/// <exception cref="AgentSpawnException">
		/// An instance of the specified agent type cannot be instantiated, because the agent type does not 
		/// provide an agent-specific constructor that accepts agent environment and agent options. Provide 
		/// a constructor with following signature: Agent(AgentEnvironment, AgentOptions).</exception>
		/// <exception cref="AgentSpawnException">
		/// An instance of the specified agent type could not be spawned. Refer to the inner exception 
		/// for more information.</exception>
		public IAgent SpawnAgent(Type agentType, AgentEnvironment environment, AgentOptions options)
		{
			Verify.IsSubclassOf(agentType, typeof(Agent), "agentType");
			Verify.NotNull(environment, "environment");
			Verify.NotNull(options, "options");

			try
			{
				var agent = OnCreate(agentType, environment, options);

				OnConnect(agent, environment, options);

				return agent;
			}
			catch (MissingMethodException exception)
			{
				throw new AgentSpawnException(
					AgentResources.AgentTypeDoesNotProvideAgentConstructor,
					exception);
			}
			catch (Exception exception)
			{
				throw new AgentSpawnException(
					AgentResources.AgentSpawnFailure,
					exception);
			}
		}

		#endregion

		#region ## Overridable Methods ##

		/// <summary>
		/// This method is called to instantiate a new agent.
		/// </summary>
		/// 
		/// <param name="agentType">
		/// The type of agent to spawn. This must be a sub-class of <see cref="Agent"/>.</param>
		/// <param name="environment">
		/// The agent environment to be supplied to the new agent instance.</param>
		/// <param name="options">
		/// The agent options to be supplied to the new agent instance.</param>
		/// 
		/// <returns>
		/// A reference to the newly spawned agent instance.</returns>
		protected virtual IAgent OnCreate(Type agentType, AgentEnvironment environment, AgentOptions options)
		{
			return (IAgent)Activator.CreateInstance(
				agentType,
				BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance,
				null,
				new Object[] { environment, options },
				null);
		}

		/// <summary>
		/// This method is called to connected the newly spawned agent to the message bus.
		/// </summary>
		/// 
		/// <param name="agent">
		/// The newly spawned agent instance.</param>
		/// <param name="environment">
		/// The agent environment to be supplied to the new agent instance.</param>
		/// <param name="options">
		/// The agent options to be supplied to the new agent instance.</param>
		protected virtual void OnConnect(IAgent agent, AgentEnvironment environment, AgentOptions options)
		{
			var runtimeEnvironment = environment as RuntimeEnvironment;

			if (runtimeEnvironment == null)
				return;

			if (MessageBusConnectedAttribute.IsDefined(agent.GetType()) == false)
				return;

			runtimeEnvironment.MessageBus.Connect(agent);
		}

		#endregion

	}
}
