﻿#region Copyright (c) 2012-11, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Reflection;


namespace Amarok.Agents
{
	/// <summary>
	/// This attribute specifies the execution mode of a message handler.
	/// </summary>
	[AttributeUsage(AttributeTargets.Method, AllowMultiple = false, Inherited = true)]
	public abstract class ExecutionModeAttribute : Attribute
	{
		#region ++ Public Static Interface ++

		/// <summary>
		/// Returns the execution mode that is defined for the given method, or the specified default value 
		/// if no execution mode is defined.
		/// </summary>
		/// 
		/// <param name="methodInfo">
		/// The method for which the execution mode should be returned.</param>
		/// <param name="defaultValue">
		/// The default execution mode that should be returned, if no execution mode is defined for the 
		/// given method.</param>
		/// 
		/// <returns>
		/// The execution mode for the given method, if defined; otherwise the specified default value.</returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static ExecutionMode GetExecutionMode(MethodInfo methodInfo, ExecutionMode defaultValue)
		{
			Verify.NotNull(methodInfo, "methodInfo");

			var attribute = (ExecutionModeAttribute)Attribute.GetCustomAttribute(
				methodInfo,
				typeof(ExecutionModeAttribute),
				true);

			if (attribute == null)
				return defaultValue;

			return attribute.ExecutionMode;
		}

		#endregion

		#region ++ Public Interface ++

		/// <summary>
		/// Gets the execution mode defined for the message handler.
		/// </summary>
		public ExecutionMode ExecutionMode
		{
			get;
			private set;
		}

		#endregion

		#region ## Protected Interface ##

		/// <summary>
		/// Initializes a new instance.
		/// </summary>
		/// 
		/// <param name="executionMode">
		/// The execution mode for the message handler.</param>
		protected ExecutionModeAttribute(ExecutionMode executionMode)
		{
			this.ExecutionMode = executionMode;
		}

		#endregion

	}
}
