﻿#region Copyright (c) 2012-11, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;


namespace Amarok.Agents
{
	/// <summary>
	/// This type defines possible parallelism levels of message handlers. The parallelism level, also known 
	/// as degree of parallelism, specifies how many instances of the same message handler can run in parallel 
	/// at the same time.
	/// </summary>
	public static class ParallelismLevel
	{
		/// <summary>
		/// Gets a constant value of 1 specifying that the message handler cannot run in parallel. The message 
		/// handler will process all messages one by one in the order as they arrive. This parallelism level is
		/// good for message handlers where processing messages in the correct order is of importance.
		/// </summary>
		public static Int32 NoParallelism
		{
			get
			{
				return 1;
			}
		}

		/// <summary>
		/// Gets a constant value of 2 specifying that at maximum two messages can be processed at the same time. 
		/// There is no guarantee that messages will be processed in the same order as they have arrived.
		/// </summary>
		public static Int32 MinimumParallelism
		{
			get
			{
				return 2;
			}
		}

		/// <summary>
		/// Gets a value that specifies a moderate parallelism level, that is the half of the maximum parallelism 
		/// that would be possible on the current machine, which itself is based on the number of processor cores.
		/// There is no guarantee that messages will be processed in the same order as they have arrived.
		/// </summary>
		public static Int32 ModerateParallelism
		{
			get
			{
				return Math.Max(2, Environment.ProcessorCount / 2);
			}
		}

		/// <summary>
		/// Gets a value that specifies the maximum parallelism level possible on the current machine. This is 
		/// based on the number of processor cores. There is no guarantee that messages will be processed in the 
		/// same order as they have arrived.
		/// </summary>
		public static Int32 MaximumParallelism
		{
			get
			{
				return Math.Max(2, Environment.ProcessorCount);
			}
		}

	}
}
