﻿#region Copyright (c) 2013-04, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Threading;
using System.Threading.Tasks;


namespace Amarok.Agents
{
	/// <summary>
	/// This static class provides various extension methods for <see cref="IDispatcher"/>.
	/// </summary>
	public static class DispatcherExtensions
	{
		#region Post(Object, Message)

		/// <summary>
		/// Posts the message to the dispatcher.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before posting.</param>
		/// <param name="message">
		/// The message that should be posted.</param>
		/// 
		/// <returns>
		/// A boolean value indicating whether the message being posted has been scheduled for further processing, or 
		/// False, if the message has been filter, the dispatcher is completing or already completed, or no matching 
		/// message handler exists.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static Boolean Post(
			this IDispatcher dispatcher,
			Object label,
			Message message)
		{
			Verify.NotNull(label, "label");
			Verify.NotNull(message, "message");

			message.ResetLabel();
			message.Label = label;

			return dispatcher.Post(message);
		}

		#endregion


		#region PostAfter(TimeSpan, Message)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="delay">
		/// The time span to wait before posting the message.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. The Task is completed in RanToCompletion state 
		/// once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			TimeSpan delay,
			Message message)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				(Int32)delay.TotalMilliseconds,
				message,
				CancellationToken.None);
		}

		#endregion

		#region PostAfter(Int32, Message)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="millisecondsDelay">
		/// The number of milliseconds to wait before posting the message.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. The Task is completed in RanToCompletion state 
		/// once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			Int32 millisecondsDelay,
			Message message)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				millisecondsDelay,
				message,
				CancellationToken.None);
		}

		#endregion

		#region PostAfter(TimeSpan, Message, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="delay">
		/// The time span to wait before posting the message.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// <param name="cancellationToken">
		/// The cancellation token that can be used to cancel the posting of the message.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. If the cancellation token is signaled before the 
		/// specified time delay, then the Task is completed in Canceled state. Otherwise, the Task is completed in 
		/// RanToCompletion state once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		/// <exception cref="ObjectDisposedException">
		/// The provided cancellation token has already been disposed.</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			TimeSpan delay,
			Message message,
			CancellationToken cancellationToken)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				(Int32)delay.TotalMilliseconds,
				message,
				cancellationToken);
		}

		#endregion

		#region PostAfter(Int32, Message, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="millisecondsDelay">
		/// The number of milliseconds to wait before posting the message.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// <param name="cancellationToken">
		/// The cancellation token that can be used to cancel the posting of the message.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. If the cancellation token is signaled before the 
		/// specified time delay, then the Task is completed in Canceled state. Otherwise, the Task is completed in 
		/// RanToCompletion state once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		/// <exception cref="ObjectDisposedException">
		/// The provided cancellation token has already been disposed.</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			Int32 millisecondsDelay,
			Message message,
			CancellationToken cancellationToken)
		{
			Verify.NotNull(message, "message");
			Verify.NotNegative(millisecondsDelay, "millisecondsDelay");

			if (dispatcher == null)
				throw new NullReferenceException();

			if (cancellationToken.IsCancellationRequested)
				return TaskHelper.Canceled;

			if (millisecondsDelay == 0)
			{
				dispatcher.Post(message);
				return TaskHelper.Completed;
			}
			else
			{
				var delayTask = TaskHelper.Delay(
					millisecondsDelay,
					cancellationToken);

#if NET45
				var postTask = delayTask.ContinueWith(
					(_task, _message) => dispatcher.Post((Message)_message),
					message,
					TaskContinuationOptions.ExecuteSynchronously | TaskContinuationOptions.OnlyOnRanToCompletion);
#else
				var postTask = delayTask.ContinueWith(
					_task => dispatcher.Post(message),
					TaskContinuationOptions.ExecuteSynchronously | TaskContinuationOptions.OnlyOnRanToCompletion);
#endif

				return postTask;
			}
		}

		#endregion


		#region PostAfter(TimeSpan, Object, Message)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="delay">
		/// The time span to wait before posting the message.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before being dispatched.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. The Task is completed in RanToCompletion state 
		/// once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			TimeSpan delay,
			Object label,
			Message message)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				(Int32)delay.TotalMilliseconds,
				label,
				message,
				CancellationToken.None);
		}

		#endregion

		#region PostAfter(Int32, Object, Message)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="millisecondsDelay">
		/// The number of milliseconds to wait before posting the message.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before being dispatched.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. The Task is completed in RanToCompletion state 
		/// once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			Int32 millisecondsDelay,
			Object label,
			Message message)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				millisecondsDelay,
				label,
				message,
				CancellationToken.None);
		}

		#endregion

		#region PostAfter(TimeSpan, Object, Message, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="delay">
		/// The time span to wait before posting the message.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before being dispatched.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// <param name="cancellationToken">
		/// The cancellation token that can be used to cancel the posting of the message.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. If the cancellation token is signaled before the 
		/// specified time delay, then the Task is completed in Canceled state. Otherwise, the Task is completed in 
		/// RanToCompletion state once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		/// <exception cref="ObjectDisposedException">
		/// The provided cancellation token has already been disposed.</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			TimeSpan delay,
			Object label,
			Message message,
			CancellationToken cancellationToken)
		{
			return DispatcherExtensions.PostAfter(
				dispatcher,
				(Int32)delay.TotalMilliseconds,
				label,
				message,
				cancellationToken);
		}

		#endregion

		#region PostAfter(Int32, Object, Message, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher after a given time delay.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="millisecondsDelay">
		/// The number of milliseconds to wait before posting the message.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before being dispatched.</param>
		/// <param name="message">
		/// The message to be posted to the dispatcher.</param>
		/// <param name="cancellationToken">
		/// The cancellation token that can be used to cancel the posting of the message.</param>
		/// 
		/// <returns>
		/// A task that represents whether the message was posted. If the cancellation token is signaled before the 
		/// specified time delay, then the Task is completed in Canceled state. Otherwise, the Task is completed in 
		/// RanToCompletion state once the specified time delay has expired and the message has been posted.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// A value was passed to a method that did not accept it as a valid argument, because the value 
		/// must be positive (equal to or greater than zero).</exception>
		/// <exception cref="ObjectDisposedException">
		/// The provided cancellation token has already been disposed.</exception>
		public static Task PostAfter(
			this IDispatcher dispatcher,
			Int32 millisecondsDelay,
			Object label,
			Message message,
			CancellationToken cancellationToken)
		{
			Verify.NotNull(label, "label");
			Verify.NotNull(message, "message");
			Verify.NotNegative(millisecondsDelay, "millisecondsDelay");

			if (dispatcher == null)
				throw new NullReferenceException();

			if (cancellationToken.IsCancellationRequested)
				return TaskHelper.Canceled;

			if (millisecondsDelay == 0)
			{
				dispatcher.Post(label, message);
				return TaskHelper.Completed;
			}
			else
			{
				var delayTask = TaskHelper.Delay(
					millisecondsDelay,
					cancellationToken);

				var postTask = delayTask.ContinueWith(
					_task => dispatcher.Post(label, message),
					TaskContinuationOptions.ExecuteSynchronously | TaskContinuationOptions.OnlyOnRanToCompletion);

				return postTask;
			}
		}

		#endregion


		#region PostOperation(Operation)

		/// <summary>
		/// Posts the message to the dispatcher.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <typeparam name="TResult">
		/// The type of result value that is returned from the operation.</typeparam>
		/// <typeparam name="TProgress">
		/// The type of progress value that is used to report progress while the operation is running.</typeparam>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="operation">
		/// The operation message that should be posted.</param>
		/// <param name="progressAction">
		/// An optional callback that is called to handle progress values reported from the operation.</param>
		/// 
		/// <returns>
		/// A task that represents the asynchronously running operation. This task can be used to wait for completion, 
		/// to register a continuation and to access state and result of the operation.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static Task<TResult> PostOperation<TResult, TProgress>(
			this IDispatcher dispatcher,
			Operation<TResult, TProgress> operation,
			Action<TProgress> progressAction = null)
		{
			Verify.NotNull(operation, "operation");

			if (progressAction != null)
				operation.Progress += progressAction;

			dispatcher.Post(operation);

			return operation.Completion;
		}

		#endregion

		#region PostOperation(Operation, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <typeparam name="TResult">
		/// The type of result value that is returned from the operation.</typeparam>
		/// <typeparam name="TProgress">
		/// The type of progress value that is used to report progress while the operation is running.</typeparam>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="operation">
		/// The operation message that should be posted.</param>
		/// <param name="cancellationToken">
		/// An CancellationToken that can be used to cancel the operation.</param>
		/// <param name="progressAction">
		/// An optional callback that is called to handle progress values reported from the operation.</param>
		/// 
		/// <returns>
		/// A task that represents the asynchronously running operation. This task can be used to wait for completion, 
		/// to register a continuation and to access state and result of the operation.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static Task<TResult> PostOperation<TResult, TProgress>(
			this IDispatcher dispatcher,
			Operation<TResult, TProgress> operation,
			CancellationToken cancellationToken,
			Action<TProgress> progressAction = null)
		{
			Verify.NotNull(operation, "operation");

			if (cancellationToken.CanBeCanceled)
			{
				cancellationToken.Register(
					state => ((Operation)state).Cancel(),
					operation,
					false);
			}

			return DispatcherExtensions.PostOperation(
				dispatcher,
				operation,
				progressAction);
		}

		#endregion


		#region PostOperation(Object, Operation)

		/// <summary>
		/// Posts the message to the dispatcher.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <typeparam name="TResult">
		/// The type of result value that is returned from the operation.</typeparam>
		/// <typeparam name="TProgress">
		/// The type of progress value that is used to report progress while the operation is running.</typeparam>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before posting.</param>
		/// <param name="operation">
		/// The operation message that should be posted.</param>
		/// <param name="progressAction">
		/// An optional callback that is called to handle progress values reported from the operation.</param>
		/// 
		/// <returns>
		/// A task that represents the asynchronously running operation. This task can be used to wait for completion, 
		/// to register a continuation and to access state and result of the operation.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static Task<TResult> PostOperation<TResult, TProgress>(
			this IDispatcher dispatcher,
			Object label,
			Operation<TResult, TProgress> operation,
			Action<TProgress> progressAction = null)
		{
			Verify.NotNull(label, "label");
			Verify.NotNull(operation, "operation");

			operation.ResetLabel();
			operation.Label = label;

			return DispatcherExtensions.PostOperation(
				dispatcher,
				operation,
				progressAction);
		}

		#endregion

		#region PostOperation(Object, Operation, CancellationToken)

		/// <summary>
		/// Posts the message to the dispatcher.
		/// 
		/// The method stores the supplied message in the dispatcher's input message queues of interested consumers and 
		/// returns immediately, meaning it only schedules the message for further processing, but the actual processing 
		/// is done later in other threads.
		/// 
		/// Messages that the dispatcher is unable to handle, i.e. because no corresponding message handler exists, 
		/// will be dropped silently.
		/// 
		/// If there is no interested consumer at the time of posting, then the message will be dropped silently.
		/// 
		/// Messages that are posted after completing the dispatcher will be dropped silently.
		/// </summary>
		/// 
		/// <typeparam name="TResult">
		/// The type of result value that is returned from the operation.</typeparam>
		/// <typeparam name="TProgress">
		/// The type of progress value that is used to report progress while the operation is running.</typeparam>
		/// 
		/// <param name="dispatcher">
		/// The dispatcher instance on which this extension method is called on.</param>
		/// <param name="label">
		/// The label that should be assigned to the message before posting.</param>
		/// <param name="operation">
		/// The operation message that should be posted.</param>
		/// <param name="cancellationToken">
		/// An CancellationToken that can be used to cancel the operation.</param>
		/// <param name="progressAction">
		/// An optional callback that is called to handle progress values reported from the operation.</param>
		/// 
		/// <returns>
		/// A task that represents the asynchronously running operation. This task can be used to wait for completion, 
		/// to register a continuation and to access state and result of the operation.
		/// </returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		public static Task<TResult> PostOperation<TResult, TProgress>(
			this IDispatcher dispatcher,
			Object label,
			Operation<TResult, TProgress> operation,
			CancellationToken cancellationToken,
			Action<TProgress> progressAction = null)
		{
			Verify.NotNull(label, "label");
			Verify.NotNull(operation, "operation");

			operation.ResetLabel();
			operation.Label = label;

			return DispatcherExtensions.PostOperation(
				dispatcher,
				operation,
				cancellationToken,
				progressAction);
		}

		#endregion

	}
}
