#region Copyright (c) 2013-03, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Reflection;
using Amarok.Agents.Resources;


namespace Amarok.Agents
{
	/// <summary>
	/// This type implements the default factory that is capable to instantiate and load subsystems.
	/// </summary>
	public class DefaultSubsystemFactory :
		ISubsystemFactory
	{
		#region ++ ISubsystemFactory Interface ++

		/// <summary>
		/// Loads the subsystem of the specified subsystem type.
		/// </summary>
		/// 
		/// <param name="subsystemType">
		/// The type of subsystem to load. This must be a sub-class of <see cref="Subsystem"/>.</param>
		/// <param name="environment">
		/// The runtime environment to be supplied to the subsystem.</param>
		/// 
		/// <returns>
		/// A reference to the newly loaded subsystem.</returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="ArgumentException">
		/// A value was passed to a method that did not accept it as a valid argument, because 
		/// the value representing a type must be a sub-class of <see cref="Subsystem"/>.</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type cannot be instantiated, because the subsystem type does not 
		/// provide a subsystem-specific constructor that accepts a runtime environment. Provide a constructor with 
		/// following signature: Subsystem(RuntimeEnvironment).</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type could not be loaded. Refer to the inner exception for more 
		/// information.</exception>
		public ISubsystem LoadSubsystem(Type subsystemType, RuntimeEnvironment environment)
		{
			Verify.IsSubclassOf(subsystemType, typeof(Subsystem), "subsystemType");
			Verify.NotNull(environment, "environment");

			try
			{
				var subsystem = OnCreate(subsystemType, environment);

				OnLoad(subsystem, environment);

				return subsystem;
			}
			catch (MissingMethodException exception)
			{
				throw new SubsystemLoadException(
					SubsystemResources.SubsystemTypeDoesNotProvideSubsystemConstructor,
					exception);
			}
			catch (Exception exception)
			{
				throw new SubsystemLoadException(
					SubsystemResources.SubsystemLoadFailure,
					exception);
			}
		}

		#endregion

		#region ## Overridable Methods ##

		/// <summary>
		/// This method is called to instantiate a new subsystem.
		/// </summary>
		/// 
		/// <param name="subsystemType">
		/// The type of subsystem to load.</param>
		/// <param name="environment">
		/// The runtime environment to be supplied to the subsystem.</param>
		/// 
		/// <returns>
		/// A reference to the newly instantiated subsystem.</returns>
		protected virtual ISubsystem OnCreate(Type subsystemType, RuntimeEnvironment environment)
		{
			return (ISubsystem)Activator.CreateInstance(
				subsystemType,
				BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance,
				null,
				new Object[] { environment },
				null);
		}

		/// <summary>
		/// This method is called to load the new subsystem.
		/// </summary>
		/// 
		/// <param name="subsystem">
		/// The subsystem to load.</param>
		/// <param name="environment">
		/// The runtime environment to be supplied to the subsystem.</param>
		protected virtual void OnLoad(ISubsystem subsystem, RuntimeEnvironment environment)
		{
			subsystem.Load();
		}

		#endregion

	}
}
