﻿#region Copyright (c) 2013-03, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;


namespace Amarok.Agents
{
	/// <summary>
	/// This type provides extension methods for <see cref="ISubsystemFactory"/>.
	/// </summary>
	public static class SubsystemFactoryExtensions
	{
		#region LoadSubsystem<T>()

		/// <summary>
		/// Loads the subsystem of the specified subsystem type.
		/// </summary>
		/// 
		/// <typeparam name="TSubsystem">
		/// The type of subsystem to load. This must be a sub-class of <see cref="Subsystem"/>.</typeparam>
		/// 
		/// <param name="factory">
		/// The subsystem factory on which the method is called on.</param>
		/// <param name="environment">
		/// The runtime environment to be supplied to the subsystem.</param>
		/// 
		/// <returns>
		/// A reference to the newly loaded subsystem.</returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type cannot be instantiated, because the subsystem type does not 
		/// provide a subsystem-specific constructor that accepts a runtime environment. Provide a constructor with 
		/// following signature: Subsystem(RuntimeEnvironment).</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type could not be loaded. Refer to the inner exception for more 
		/// information.</exception>
		public static ISubsystem LoadSubsystem<TSubsystem>(this ISubsystemFactory factory, RuntimeEnvironment environment)
			where TSubsystem : Subsystem
		{
			return factory.LoadSubsystem(
				typeof(TSubsystem),
				environment);
		}

		#endregion

		#region LoadSubsystem<T, E>()

		/// <summary>
		/// Loads the subsystem of the specified subsystem type.
		/// </summary>
		/// 
		/// <typeparam name="TSubsystem">
		/// The type of subsystem to load. This must be a sub-class of <see cref="Subsystem"/>.</typeparam>
		/// <typeparam name="TEnvironment">
		/// The type of runtime environment to supply. This must be a sub-class of <see cref="RuntimeEnvironment"/>.</typeparam>
		/// 
		/// <param name="factory">
		/// The subsystem factory on which the method is called on.</param>
		/// <param name="environment">
		/// The runtime environment to be supplied to the subsystem.</param>
		/// 
		/// <returns>
		/// A reference to the newly loaded subsystem.</returns>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type cannot be instantiated, because the subsystem type does not 
		/// provide a subsystem-specific constructor that accepts a runtime environment. Provide a constructor with 
		/// following signature: Subsystem(RuntimeEnvironment).</exception>
		/// <exception cref="SubsystemLoadException">
		/// An instance of the specified subsystem type could not be loaded. Refer to the inner exception for more 
		/// information.</exception>
		public static ISubsystem LoadSubsystem<TSubsystem, TEnvironment>(this ISubsystemFactory factory, TEnvironment environment)
			where TSubsystem : Subsystem
			where TEnvironment : RuntimeEnvironment
		{
			return factory.LoadSubsystem(
				typeof(TSubsystem),
				environment);
		}

		#endregion

	}
}
