﻿#region Copyright (c) 2012-10, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;


namespace Amarok
{
	/// <summary>
	/// This type represents a base class for Entity implementations.
	/// </summary>
	[Serializable]
	public abstract class Entity
	{
		// references
		private readonly Identity mIdentity;


		#region ++ Public Interface ++

		/// <summary>
		/// Gets a reference to the identity of this entity.
		/// </summary>
		public Identity Identity
		{
			get
			{
				return mIdentity;
			}
		}

		/// <summary>
		/// Gets a reference to the identity of the supplied entity.
		/// </summary>
		/// 
		/// <param name="entity">
		/// The entity from which to return the identity.</param>
		/// 
		/// <returns>
		/// The identity of the supplied entity, or null if null was supplied.</returns>
		public static implicit operator Identity(Entity entity)
		{
			if (entity == null)
				return null;

			return entity.Identity;
		}

		/// <summary>
		/// Returns a reference to the identity of this entity.
		/// </summary>
		/// 
		/// <returns>
		/// The identity of this entity.</returns>
		public Identity ToIdentity()
		{
			return mIdentity;
		}

		#endregion

		#region ## Protected Interface ##

		/// <summary>
		/// Initializes a new instance.
		/// </summary>
		/// 
		/// <param name="identity">
		/// The identity if this entity.</param>
		/// 
		/// <exception cref="ArgumentNullException">
		/// A null reference was passed to a method that did not accept it as a valid argument.</exception>
		protected Entity(Identity identity)
		{
			Verify.NotNull(identity, "identity");
			mIdentity = identity;
		}

		#endregion

	}
}
