﻿#region Copyright (c) 2013-02, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;


namespace Amarok
{
	/// <summary>
	/// This base class manages an object-id that consists of the object's type-name and an optional instance-name. 
	/// The object's type-name is constant. In contrast, the object's instance-name can be set and changed during 
	/// the object's lifetime. The resulting object-id, a String formatted as "type-name[:instance-name]", can be 
	/// used to distinguish one object instance from other object instances.
	/// </summary>
	public abstract class IdentifiableObject :
		IIdentifiableObject
	{
		// data
		private String mObjectId;


		#region ++ IIdentifiableObject Interface ++

		/// <summary>
		/// Gets a String formatted as "type-name[:instance-name]" that represents the object-id of this object 
		/// instance. The instance-name is null by default, but can be changed throughout the object's lifetime.
		/// </summary>
		String IIdentifiableObject.ObjectId
		{
			get
			{
				return mObjectId;
			}
		}

		#endregion

		#region ++ Public Interface ++

		/// <summary>
		/// Returns a String formatted as "type-name[:instance-name]" that represents the object-id of this object 
		/// instance. The instance-name is null by default, but can be changed throughout the object's lifetime.
		/// </summary>
		public override String ToString()
		{
			return mObjectId;
		}

		#endregion

		#region ## Protected Interface ##

		/// <summary>
		/// Gets a String formatted as "type-name[:instance-name]" that represents the object-id of this object 
		/// instance. The instance-name is null by default, but can be changed throughout the object's lifetime.
		/// </summary>
		protected String ObjectId
		{
			get
			{
				return mObjectId;
			}
		}


		/// <summary>
		/// Initializes a new instance. Sets the object-id to "type-name".
		/// </summary>
		protected IdentifiableObject()
		{
			mObjectId = _FormatTypeName(this);
		}

		/// <summary>
		/// Initializes a new instance. Sets the object-id to "type-name[:instance-name]".
		/// </summary>
		/// 
		/// <param name="instanceName">
		/// The instance-name of this object. Might be null or an empty string.</param>
		protected IdentifiableObject(String instanceName)
		{
			mObjectId = _FormatObjectId(
				_FormatTypeName(this),
				instanceName);
		}


		/// <summary>
		/// Changes the instance-name of this object.
		/// 
		/// Warning: This method can be used in multi-threaded scenarios, but your latest value won't be visible 
		/// immediately to other threads. It is not recommended to call this method concurrently from multiple 
		/// threads.
		/// </summary>
		/// 
		/// <param name="instanceName">
		/// The instance-name of this object. Might be null or an empty string.</param>
		protected virtual void SetInstanceName(String instanceName)
		{
			mObjectId = _FormatObjectId(
				_FormatTypeName(this),
				instanceName);
		}

		#endregion

		#region Implementation

		private static String _FormatTypeName(Object objectInstance)
		{
			return objectInstance.GetType().FullName;
		}

		private static String _FormatObjectId(String typeName, String instanceName)
		{
			if (String.IsNullOrEmpty(instanceName))
				return typeName;

			return typeName + ':' + instanceName;
		}

		#endregion

	}
}
