﻿#region Copyright (c) 2012-10, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System.Threading.Tasks;


namespace Amarok
{
	/// <summary>
	/// This type represents a <see cref="TaskCompletionSource{T}"/> that can just signal completion. The associated
	/// Task cannot complete in a faulted or canceled state and it does not return any completion result. The sole 
	/// purpose of this Task is signaling completion.
	/// </summary>
	public sealed class TaskCompletionSource
	{
		// data
		private readonly TaskCompletionSource<None> mSource = new TaskCompletionSource<None>();


		#region ++ Public Interface ++

		/// <summary>
		/// Gets a reference to the Task associated with this TaskCompletionSource. This Task instance will 
		/// transition into the final RunToCompletion state to signal completion.
		/// </summary>
		public Task Task
		{
			get
			{
				return mSource.Task;
			}
		}


		/// <summary>
		/// Transitions the Task into the final RunToCompletion state to signal completion. It is safe to call
		/// this method twice. The method will return immediately without throwing any exception.
		/// </summary>
		public void SignalCompletion()
		{
			mSource.TrySetResult(None.Instance);
		}

		#endregion

		#region (Testing)
#if TESTS

		/// <summary>
		/// !!! For testing purposes only !!!
		/// </summary>
		public TaskCompletionSource<None> InnerTaskCompletionSource
		{
			get
			{
				return mSource;
			}
		}

#endif
		#endregion

	}
}
