﻿#region Copyright (c) 07/17/2012, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Globalization;
using System.Windows;
using System.Windows.Data;


namespace Amarok.Presentation
{
	/// <summary>
	/// Converts the given Boolean value to a configurable Object value.
	/// 
	/// :: Convert
	/// True   ...  TrueValue property
	/// False  ...  FalseValue property
	/// 
	/// :: ConvertBack
	/// equal to TrueValue   ...  True
	/// equal to FalseValue  ...  False
	/// Otherwise            ...  DefaultValue
	/// </summary>
	[ValueConversion(typeof(Boolean), typeof(Object))]
	public sealed class BooleanToObjectConverter : ValueConverter<Boolean, Object>
	{
		/// <summary>
		/// Gets or sets the Object that is returned by Convert() when True is supplied. Defaults to True.
		/// </summary>
		public Object TrueValue
		{
			get;
			set;
		}

		/// <summary>
		/// Gets or sets the Object that is returned by Convert() when False is supplied. Defaults to False.
		/// </summary>
		public Object FalseValue
		{
			get;
			set;
		}

		/// <summary>
		/// Gets or sets the Object that is returned by ConvertBack() when neither TrueValue nor FalseValue 
		/// are equal to the supplied value. Defaults to DependencyProperty.UnsetValue.
		/// </summary>
		public Object DefaultValue
		{
			get;
			set;
		}


		/// <summary>
		/// Initializes a new instance.
		/// </summary>
		public BooleanToObjectConverter()
		{
			this.TrueValue = true;
			this.FalseValue = false;
			this.DefaultValue = DependencyProperty.UnsetValue;
		}


		/// <summary>
		/// </summary>
		protected override Object OnConvert(Boolean value, Type targetType, Object parameter, CultureInfo culture)
		{
			return value ?
				this.TrueValue :
				this.FalseValue;
		}

		/// <summary>
		/// </summary>
		protected override Object OnConvertBack(Object value, Type targetType, Object parameter, CultureInfo culture)
		{
			if (Object.Equals(value, this.TrueValue))
				return true;

			if (Object.Equals(value, this.FalseValue))
				return false;

			return this.DefaultValue;
		}

	}
}
