﻿#region Copyright (c) 2013-08, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

using System;
using System.Collections.Specialized;


namespace Amarok.Presentation
{
	/// <summary>
	/// This type provides hooks (events) for diagnosis purposes.
	/// </summary>
	public static class ViewModelHooks
	{
		#region ++ Public Static Interface (Events) ++

		/// <summary>
		/// Occurs when a property changed event on a view model object was raised.
		/// </summary>
		public static event EventHandler<PropertyChangedEventArgs> PropertyChanged;

		/// <summary>
		/// Occurs when a collection changed event on a view model collection object was raised.
		/// </summary>
		public static event EventHandler<CollectionChangedEventArgs> CollectionChanged;


		/// <summary>
		/// Removes all event handler
		/// </summary>
		public static void Reset()
		{
			PropertyChanged = null;
			CollectionChanged = null;
		}

		#endregion

		#region ~~ Internal Static Interface ~~

		/// <summary>
		/// </summary>
		internal static void RaisePropertyChanged(Object viewModel, String propertyName)
		{
			try
			{
				var handler = PropertyChanged;

				if (handler != null)
				{
					handler(null, new PropertyChangedEventArgs(viewModel, propertyName));
				}
			}
			catch (Exception)
			{
				// ignore
			}
		}

		/// <summary>
		/// </summary>
		internal static void RaiseCollectionChanged(Object viewModel, NotifyCollectionChangedAction action, Int32 newStartingIndex, Int32 oldStartingIndex)
		{
			try
			{
				var handler = CollectionChanged;

				if (handler != null)
				{
					handler(null, new CollectionChangedEventArgs(viewModel, action, newStartingIndex, oldStartingIndex));
				}
			}
			catch (Exception)
			{
				// ignore
			}
		}

		#endregion

		#region ++ Public Static Interface (Event Args) ++

		/// <summary>
		/// </summary>
		public abstract class ViewModelHooksEventArgs : EventArgs
		{
			/// <summary>
			/// </summary>
			public DateTime TimestampUtc
			{
				get;
				private set;
			}

			/// <summary>
			/// </summary>
			public Object ViewModel
			{
				get;
				private set;
			}


			/// <summary>
			/// </summary>
			protected ViewModelHooksEventArgs(Object viewModel)
			{
				this.TimestampUtc = DateTime.UtcNow;
				this.ViewModel = viewModel;
			}
		}

		/// <summary>
		/// </summary>
		public sealed class PropertyChangedEventArgs : ViewModelHooksEventArgs
		{
			/// <summary>
			/// </summary>
			public String PropertyName
			{
				get;
				private set;
			}


			/// <summary>
			/// </summary>
			public PropertyChangedEventArgs(Object viewModel, String propertyName)
				: base(viewModel)
			{
				this.PropertyName = propertyName;
			}
		}

		/// <summary>
		/// </summary>
		public sealed class CollectionChangedEventArgs : ViewModelHooksEventArgs
		{
			/// <summary>
			/// </summary>
			public NotifyCollectionChangedAction Action
			{
				get;
				private set;
			}

			/// <summary>
			/// </summary>
			public Int32 NewStartingIndex
			{
				get;
				private set;
			}

			/// <summary>
			/// </summary>
			public Int32 OldStartingIndex
			{
				get;
				private set;
			}


			/// <summary>
			/// </summary>
			public CollectionChangedEventArgs(Object viewModel, NotifyCollectionChangedAction action, Int32 newStartingIndex, Int32 oldStartingIndex)
				: base(viewModel)
			{
				this.Action = action;
				this.NewStartingIndex = newStartingIndex;
				this.OldStartingIndex = oldStartingIndex;
			}
		}

		#endregion

	}
}
