﻿#region Copyright (c) 2012-11, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

#if TESTS

using System;
using NUnit.Framework;


namespace Amarok.Agents
{
	[TestFixture]
	public class Test_ExecutionModeAttributes
	{
		private abstract class MyBaseClass
		{
			[ConcurrentExecution]
			public abstract void TestMethod5();

			public abstract void TestMethod6();
		}

		private class MyClass : MyBaseClass
		{
			public void TestMethod1()
			{
			}

			[IndependentExecution]
			public void TestMethod2()
			{
			}

			[ExclusiveExecution]
			public void TestMethod3()
			{
			}

			[ConcurrentExecution]
			public void TestMethod4()
			{
			}

			public override void TestMethod5()
			{
			}

			[ExclusiveExecution]
			public override void TestMethod6()
			{
			}
		}


		// --- TESTS ---


		[Test]
		public void Test_GetExecutionMode_NoAttribute()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod1");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Exclusive),
				Is.EqualTo(ExecutionMode.Exclusive));
		}

		[Test]
		public void Test_GetExecutionMode_IndependentAttribute()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod2");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Exclusive),
				Is.EqualTo(ExecutionMode.Independent));
		}

		[Test]
		public void Test_GetExecutionMode_ExclusiveAttribute()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod3");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Concurrent),
				Is.EqualTo(ExecutionMode.Exclusive));
		}

		[Test]
		public void Test_GetExecutionMode_ConcurrentAttribute()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod4");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Independent),
				Is.EqualTo(ExecutionMode.Concurrent));
		}

		[Test]
		public void Test_GetExecutionMode_AbstractMethod_AttributeOnBaseMethod()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod5");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Independent),
				Is.EqualTo(ExecutionMode.Concurrent));
		}

		[Test]
		public void Test_GetExecutionMode_AbstractMethod_AttributeOnOverridingMethod()
		{
			var target = new MyClass();
			var targetType = target.GetType();
			var method = targetType.GetMethod("TestMethod6");

			Assert.That(method, Is.Not.Null);
			Assert.That(ExecutionModeAttribute.GetExecutionMode(method, ExecutionMode.Independent),
				Is.EqualTo(ExecutionMode.Exclusive));
		}

		[Test]
		public void Test_GetExecutionMode_NullMethod()
		{
			Assert.Throws<ArgumentNullException>(() =>
				{
					ExecutionModeAttribute.GetExecutionMode(null, ExecutionMode.Independent);
				});
		}

	}
}

#endif
