﻿#region Copyright (c) 2013-04, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

#if TESTS

using System;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;


namespace Amarok.Agents
{
	[TestFixture]
	public class Test_OperationHandlerInfo
	{
		private class MyOperation : Operation<String, Int32>
		{
		}

		private class MyClass
		{
			[IndependentExecution, MaximumParallelismAttribute]
			public String Method1(String argument)
			{
				return null;
			}

			public String Method2(Int32 argument)
			{
				return null;
			}


			public static MyOperation Argument;
			public static Int32 Called;

			[ExclusiveExecution, NoParallelismAttribute]
			public String Method3(MyOperation argument)
			{
				Argument = argument;
				Called++;
				return "done";
			}

			[ExclusiveExecution, NoParallelismAttribute]
			public static String Method4(MyOperation argument)
			{
				Argument = argument;
				Called++;
				return "done";
			}

			[ExclusiveExecution, NoParallelismAttribute]
			public async Task<String> Method5(MyOperation argument)
			{
				Argument = argument;
				Called++;

				await Task.Factory.StartNew(() => Thread.Sleep(250));

				return "done";
			}
		}


		// --- TESTS ---


		[Test]
		public void Test_Construction()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method1");
			var parameterInfo = methodInfo.GetParameters()[0];

			var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
				typeof(String), typeof(Int32));

			Assert.That(handlerInfo.MethodInfo, Is.EqualTo(methodInfo));
			Assert.That(handlerInfo.ParameterInfo, Is.EqualTo(parameterInfo));
			Assert.That(handlerInfo.IsAsync, Is.False);
			Assert.That(handlerInfo.ResultType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ProgressType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.MessageType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ExecutionMode, Is.EqualTo(ExecutionMode.Independent));
			Assert.That(handlerInfo.ParallelismLevel, Is.EqualTo(ParallelismLevel.MaximumParallelism));
		}

		[Test]
		public void Test_Construction_Defaults()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method2");
			var parameterInfo = methodInfo.GetParameters()[0];

			var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
				typeof(String), typeof(Int32));

			Assert.That(handlerInfo.MethodInfo, Is.EqualTo(methodInfo));
			Assert.That(handlerInfo.ParameterInfo, Is.EqualTo(parameterInfo));
			Assert.That(handlerInfo.IsAsync, Is.False);
			Assert.That(handlerInfo.ResultType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ProgressType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.MessageType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.ExecutionMode, Is.EqualTo(ExecutionMode.Exclusive));
			Assert.That(handlerInfo.ParallelismLevel, Is.EqualTo(ParallelismLevel.NoParallelism));
		}

		[Test]
		public void Test_CreateHandler_InvokeHandler()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method3");
			var parameterInfo = methodInfo.GetParameters()[0];

			var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
				typeof(String), typeof(Int32));

			Assert.That(handlerInfo.MethodInfo, Is.EqualTo(methodInfo));
			Assert.That(handlerInfo.ParameterInfo, Is.EqualTo(parameterInfo));
			Assert.That(handlerInfo.IsAsync, Is.False);
			Assert.That(handlerInfo.ResultType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ProgressType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.MessageType, Is.EqualTo(typeof(MyOperation)));
			Assert.That(handlerInfo.ExecutionMode, Is.EqualTo(ExecutionMode.Exclusive));
			Assert.That(handlerInfo.ParallelismLevel, Is.EqualTo(ParallelismLevel.NoParallelism));

			MyClass.Argument = null;
			MyClass.Called = 0;

			var handler = handlerInfo.CreateHandler(target);

			Assert.That(MyClass.Called, Is.EqualTo(0));
			Assert.That(MyClass.Argument, Is.Null);

			var message1 = new MyOperation();
			handler.Invoke(message1);

			Assert.That(MyClass.Called, Is.EqualTo(1));
			Assert.That(MyClass.Argument, Is.SameAs(message1));

			var message2 = new MyOperation();
			handler.Invoke(message2);

			Assert.That(MyClass.Called, Is.EqualTo(2));
			Assert.That(MyClass.Argument, Is.SameAs(message2));
		}

		[Test]
		public void Test_CreateHandler_InvokeHandler_StaticMethod()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method4");
			var parameterInfo = methodInfo.GetParameters()[0];

			var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
				typeof(String), typeof(Int32));

			Assert.That(handlerInfo.MethodInfo, Is.EqualTo(methodInfo));
			Assert.That(handlerInfo.ParameterInfo, Is.EqualTo(parameterInfo));
			Assert.That(handlerInfo.IsAsync, Is.False);
			Assert.That(handlerInfo.ResultType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ProgressType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.MessageType, Is.EqualTo(typeof(MyOperation)));
			Assert.That(handlerInfo.ExecutionMode, Is.EqualTo(ExecutionMode.Exclusive));
			Assert.That(handlerInfo.ParallelismLevel, Is.EqualTo(ParallelismLevel.NoParallelism));

			MyClass.Argument = null;
			MyClass.Called = 0;

			var handler = handlerInfo.CreateHandler(null);

			Assert.That(MyClass.Called, Is.EqualTo(0));
			Assert.That(MyClass.Argument, Is.Null);

			var message = new MyOperation();
			handler.Invoke(message);

			Assert.That(MyClass.Called, Is.EqualTo(1));
			Assert.That(MyClass.Argument, Is.SameAs(message));
		}

		[Test]
		public void Test_CreateHandler_InvokeHandler_Async()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method5");
			var parameterInfo = methodInfo.GetParameters()[0];

			var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, true,
				typeof(String), typeof(Int32));

			Assert.That(handlerInfo.MethodInfo, Is.EqualTo(methodInfo));
			Assert.That(handlerInfo.ParameterInfo, Is.EqualTo(parameterInfo));
			Assert.That(handlerInfo.IsAsync, Is.True);
			Assert.That(handlerInfo.ResultType, Is.EqualTo(typeof(String)));
			Assert.That(handlerInfo.ProgressType, Is.EqualTo(typeof(Int32)));
			Assert.That(handlerInfo.MessageType, Is.EqualTo(typeof(MyOperation)));
			Assert.That(handlerInfo.ExecutionMode, Is.EqualTo(ExecutionMode.Exclusive));
			Assert.That(handlerInfo.ParallelismLevel, Is.EqualTo(ParallelismLevel.NoParallelism));

			MyClass.Argument = null;
			MyClass.Called = 0;

			var handler = handlerInfo.CreateHandler(target);

			Assert.That(MyClass.Called, Is.EqualTo(0));
			Assert.That(MyClass.Argument, Is.Null);

			var message1 = new MyOperation();
			handler.Invoke(message1);
			message1.Completion.WaitQuietly();

			Assert.That(MyClass.Called, Is.EqualTo(1));
			Assert.That(MyClass.Argument, Is.SameAs(message1));

			var message2 = new MyOperation();
			handler.Invoke(message2);
			message2.Completion.WaitQuietly();

			Assert.That(MyClass.Called, Is.EqualTo(2));
			Assert.That(MyClass.Argument, Is.SameAs(message2));
		}

		[Test]
		public void Test_Construction_NullMethodInfo()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method1");
			var parameterInfo = methodInfo.GetParameters()[0];

			Assert.Throws<ArgumentNullException>(() =>
				{
					var handlerInfo = new OperationHandlerInfo(null, parameterInfo, false,
						typeof(String), typeof(Int32));
				});
		}

		[Test]
		public void Test_Construction_NullParameterInfo()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method1");

			Assert.Throws<ArgumentNullException>(() =>
				{
					var handlerInfo = new OperationHandlerInfo(methodInfo, null, false,
						typeof(String), typeof(Int32));
				});
		}

		[Test]
		public void Test_Construction_NullOperationResultType()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method1");
			var parameterInfo = methodInfo.GetParameters()[0];

			Assert.Throws<ArgumentNullException>(() =>
				{
					var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
						null, typeof(Int32));
				});
		}

		[Test]
		public void Test_Construction_NullOperationProgressType()
		{
			var target = new MyClass();
			var methodInfo = target.GetType().GetMethod("Method1");
			var parameterInfo = methodInfo.GetParameters()[0];

			Assert.Throws<ArgumentNullException>(() =>
				{
					var handlerInfo = new OperationHandlerInfo(methodInfo, parameterInfo, false,
						typeof(String), null);
				});
		}

	}
}

#endif
