﻿#region Copyright (c) 2012-10, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

#if TESTS

using System;
using NUnit.Framework;


namespace Amarok
{
	[TestFixture]
	public class Test_DisposableObject
	{
		private class TestDisposableObject : DisposableObject
		{
			public Int32 OnDisposeCalled;

			public TestDisposableObject()
			{
			}

			public TestDisposableObject(String instanceName)
				: base(instanceName)
			{
			}

			public Boolean CallBaseIsDisposed()
			{
				return base.IsDisposed;
			}

			public void CallBaseThrowIfDisposed()
			{
				base.ThrowIfDisposed();
			}

			public void CallBaseDisposeObject()
			{
				base.DisposeObject();
				Assert.That(base.IsDisposed, Is.True);
			}

			public void CallBaseSetInstanceName(String instanceName)
			{
				base.SetInstanceName(instanceName);
			}

			public String CallBaseGetObjectId()
			{
				return base.ObjectId;
			}

			protected override void OnDispose()
			{
				base.OnDispose();

				Assert.That(base.IsDisposed, Is.True);
				OnDisposeCalled++;
			}
		}


		// --- TESTS ---


		[Test]
		public void Test_Construction()
		{
			var obj = new TestDisposableObject();

			Assert.That(obj, Is.InstanceOf<IDisposable>());
			Assert.That(obj.CallBaseIsDisposed(), Is.False);
			Assert.DoesNotThrow(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(0));
		}

		[Test]
		public void Test_Construction_InstanceName()
		{
			var obj = new TestDisposableObject("INSTANCE");

			Assert.That(obj, Is.InstanceOf<IDisposable>());
			Assert.That(obj.CallBaseIsDisposed(), Is.False);
			Assert.DoesNotThrow(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(0));
			Assert.That(obj.CallBaseGetObjectId(), Is.StringContaining("INSTANCE"));
		}

		[Test]
		public void Test_Dispose()
		{
			var obj = new TestDisposableObject();
			((IDisposable)obj).Dispose();

			Assert.That(obj.CallBaseIsDisposed(), Is.True);
			var exception = Assert.Throws<ObjectDisposedException>(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(1));
			Assert.That(exception.ObjectName, Is.EqualTo(obj.CallBaseGetObjectId()));
		}

		[Test]
		public void Test_Dispose_InstanceName()
		{
			var obj = new TestDisposableObject();
			obj.CallBaseSetInstanceName("TEST");

			((IDisposable)obj).Dispose();

			Assert.That(obj.CallBaseIsDisposed(), Is.True);
			var exception = Assert.Throws<ObjectDisposedException>(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(1));
			Assert.That(exception.ObjectName, Is.EqualTo(obj.CallBaseGetObjectId()));
			Assert.That(exception.ObjectName, Is.StringContaining("TEST"));
		}

		[Test]
		public void Test_Dispose_Twice()
		{
			var obj = new TestDisposableObject();
			((IDisposable)obj).Dispose();
			((IDisposable)obj).Dispose();

			Assert.That(obj.CallBaseIsDisposed(), Is.True);
			Assert.Throws<ObjectDisposedException>(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(1));
		}

		[Test]
		public void Test_DisposeObject()
		{
			var obj = new TestDisposableObject();
			obj.CallBaseDisposeObject();

			Assert.That(obj.CallBaseIsDisposed(), Is.True);
			Assert.Throws<ObjectDisposedException>(() => obj.CallBaseThrowIfDisposed());
			Assert.That(obj.OnDisposeCalled, Is.EqualTo(1));
		}

	}
}

#endif
