﻿#region Copyright (c) 2012-10, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

#if TESTS

using System;
using System.Threading;
using NUnit.Framework;


namespace Amarok
{
	[TestFixture]
	public class Test_IntervalConstrainedProgress
	{
		private class TestProgress : IProgress<Int32>
		{
			public Int32 LastValue = -1;

			public void Report(Int32 value)
			{
				LastValue = value;
			}
		}


		// --- TESTS ---


		[Test]
		public void Test_Construction()
		{
			var testProgress = new TestProgress();
			var progress = new IntervalConstrainedProgress<Int32>(testProgress, TimeSpan.FromMilliseconds(500));

			Assert.That(progress.Progress, Is.SameAs(testProgress));
			Assert.That(progress.MinimumInterval, Is.EqualTo(TimeSpan.FromMilliseconds(500)));
		}

		[Test]
		public void Test_Construction_NullProgress()
		{
			Assert.Throws<ArgumentNullException>(() =>
				{
					var progress = new IntervalConstrainedProgress<Int32>(null, TimeSpan.FromMilliseconds(500));
				});
		}

		[Test]
		public void Test_Construction_NegativeInterval()
		{
			Assert.Throws<ArgumentOutOfRangeException>(() =>
				{
					var testProgress = new TestProgress();
					var progress = new IntervalConstrainedProgress<Int32>(testProgress, TimeSpan.FromMilliseconds(-10));
				});
		}

		[Test]
		public void Test_Report()
		{
			var testProgress = new TestProgress();
			var progress = new IntervalConstrainedProgress<Int32>(testProgress, TimeSpan.FromMilliseconds(50));

			Assert.That(testProgress.LastValue, Is.EqualTo(-1));

			// first in interval (always)
			progress.Report(10);
			Assert.That(testProgress.LastValue, Is.EqualTo(10));

			// second in interval (ignored)
			progress.Report(20);
			Assert.That(testProgress.LastValue, Is.EqualTo(10));

			// third in second interval
			Thread.Sleep(250);
			progress.Report(30);
			Assert.That(testProgress.LastValue, Is.EqualTo(30));

			// fourth in second interval (ignored)
			progress.Report(40);
			Assert.That(testProgress.LastValue, Is.EqualTo(30));

			// fifth in third interval
			Thread.Sleep(250);
			progress.Report(50);
			Assert.That(testProgress.LastValue, Is.EqualTo(50));
		}

		[Test]
		public void Test_Report_ValueFunc()
		{
			var testProgress = new TestProgress();
			var progress = new IntervalConstrainedProgress<Int32>(testProgress, TimeSpan.FromMilliseconds(50));

			Assert.That(testProgress.LastValue, Is.EqualTo(-1));

			// first in interval (always)
			progress.Report(() => 10);
			Assert.That(testProgress.LastValue, Is.EqualTo(10));

			// second in interval (ignored)
			progress.Report(() => 20);
			Assert.That(testProgress.LastValue, Is.EqualTo(10));

			// third in second interval
			Thread.Sleep(250);
			progress.Report(() => 30);
			Assert.That(testProgress.LastValue, Is.EqualTo(30));

			// fourth in second interval (ignored)
			progress.Report(() => 40);
			Assert.That(testProgress.LastValue, Is.EqualTo(30));

			// fifth in third interval
			Thread.Sleep(250);
			progress.Report(() => 50);
			Assert.That(testProgress.LastValue, Is.EqualTo(50));
		}

		[Test]
		public void Test_Report_NullValueFunc()
		{
			var testProgress = new TestProgress();
			var progress = new IntervalConstrainedProgress<Int32>(testProgress, TimeSpan.FromMilliseconds(50));

			Assert.Throws<ArgumentNullException>(() =>
				{
					progress.Report(null);
				});
		}

	}
}

#endif
