﻿#region Copyright (c) 2013-05, Olaf Kober <amarok.projects@gmail.com>
//================================================================================
//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.
//================================================================================
#endregion

#if TESTS

using System;
using NUnit.Framework;


namespace Amarok.Presentation
{
	[TestFixture]
	public class Test_ViewCommand
	{
		[Test]
		public void Test_Usage_Execute()
		{
			var executeCalled = 0;

			var command = new ViewCommand(
				() =>
				{
					executeCalled++;
				});

			Assert.That(executeCalled, Is.EqualTo(0));

			command.Execute(null);

			Assert.That(executeCalled, Is.EqualTo(1));

			var result = command.CanExecute(null);

			Assert.That(result, Is.True);

			command.Execute("BLAH");

			Assert.That(executeCalled, Is.EqualTo(2));
		}

		[Test]
		public void Test_Usage_CanExecute()
		{
			var canExecuteCalled = 0;
			var executeCalled = 0;
			var canExecuteResult = true;

			var command = new ViewCommand(
				() =>
				{
					executeCalled++;
				},
				() =>
				{
					canExecuteCalled++;
					return canExecuteResult;
				});

			Assert.That(executeCalled, Is.EqualTo(0));
			Assert.That(canExecuteCalled, Is.EqualTo(0));

			command.CanExecute(null);

			Assert.That(executeCalled, Is.EqualTo(0));
			Assert.That(canExecuteCalled, Is.EqualTo(1));

			command.Execute(null);

			Assert.That(executeCalled, Is.EqualTo(1));
			Assert.That(canExecuteCalled, Is.EqualTo(1));

			command.CanExecute("BLAH");

			Assert.That(executeCalled, Is.EqualTo(1));
			Assert.That(canExecuteCalled, Is.EqualTo(2));

			command.Execute("BLAH");

			Assert.That(executeCalled, Is.EqualTo(2));
			Assert.That(canExecuteCalled, Is.EqualTo(2));

			canExecuteResult = false;
			command.Execute("BLAH");

			Assert.That(executeCalled, Is.EqualTo(3));
			Assert.That(canExecuteCalled, Is.EqualTo(2));
		}

		[Test]
		public void Test_Usage_Execute_After_Dispose()
		{
			var executeCalled = 0;

			var command = new ViewCommand(
				() =>
				{
					executeCalled++;
				});

			Assert.That(command.IsDisposed, Is.False);
			Assert.That(executeCalled, Is.EqualTo(0));

			command.Execute(null);

			Assert.That(command.IsDisposed, Is.False);
			Assert.That(executeCalled, Is.EqualTo(1));

			Helper.Dispose(command);
			command.Execute(null);

			Assert.That(command.IsDisposed, Is.True);
			Assert.That(executeCalled, Is.EqualTo(1));
		}

		[Test]
		public void Test_Usage_CanExecute_After_Dispose()
		{
			var canExecuteCalled = 0;
			var executeCalled = 0;
			var canExecuteResult = true;

			var command = new ViewCommand(
				() =>
				{
					executeCalled++;
				},
				() =>
				{
					canExecuteCalled++;
					return canExecuteResult;
				});

			Assert.That(command.IsDisposed, Is.False);
			Assert.That(executeCalled, Is.EqualTo(0));
			Assert.That(canExecuteCalled, Is.EqualTo(0));

			var result1 = command.CanExecute(null);

			Assert.That(command.IsDisposed, Is.False);
			Assert.That(executeCalled, Is.EqualTo(0));
			Assert.That(canExecuteCalled, Is.EqualTo(1));
			Assert.That(result1, Is.True);

			Helper.Dispose(command);
			var result2 = command.CanExecute(null);

			Assert.That(command.IsDisposed, Is.True);
			Assert.That(executeCalled, Is.EqualTo(0));
			Assert.That(canExecuteCalled, Is.EqualTo(1));
			Assert.That(result2, Is.False);
		}

		[Test]
		public void Test_Usage_CanExecuteChanged_Without_CanExecuteHandler()
		{
			var command = new ViewCommand(
				() =>
				{
				});

			var handlerCalled = 0;
			EventHandler handler = (sender, e) =>
				handlerCalled++;

			command.CanExecuteChanged += handler;
			command.RaiseCanExecuteChanged();

			//Assert.That(handlerCalled, Is.EqualTo(0));		// DOES NOT WORK WITHOUT UI

			command.CanExecuteChanged -= handler;
		}

		[Test]
		public void Test_Usage_CanExecuteChanged_With_CanExecuteHandler()
		{
			var command = new ViewCommand(
				() =>
				{
				},
				() =>
				{
					return false;
				});

			Int32 handlerCalled = 0;
			EventHandler handler = (sender, e) =>
				handlerCalled++;

			command.CanExecuteChanged += handler;
			command.RaiseCanExecuteChanged();

			//Assert.That(handlerCalled, Is.EqualTo(1));		// DOES NOT WORK WITHOUT UI

			command.CanExecuteChanged -= handler;
		}

	}
}

#endif
